import logging
import networkx as nx
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import pygna.output as out
import pygna.plots as pp
import seaborn as sns
import math
import pygna.reading_class as rc
from palettable.colorbrewer.diverging import *
from palettable.colorbrewer.sequential import *


#######################################################
######## COMMAND FUNCTIONS ############################
########################################################


def paint_datasets_stats(table_filename: 'pygna results table',
                         output_file: 'figure file, use pdf or png extension',
                         alternative="greater"
                         ):
    """
    This function plots the results of of a GNT test.
    Pass the results table generated by one of the functions and the output figure file (png or pdf).
    In case you are using a SP test, pass also 'less' as an alternative.
    """

    palette_binary = RdBu_4.mpl_colors[0::3]

    with open(table_filename, "r") as f:
        table = pd.read_csv(f, sep=",")

    stat_name = table["analysis"][0]
    n_permutations = table["number_of_permutations"][0]
    data = pd.DataFrame()

    # If greater the test is between 0 and + inf
    # else is bwteen -inf and 0, I need to set the right bounds
    if alternative == "greater":
        for i, row in table.sort_values(
            by=["empirical_pvalue", "observed"], ascending=[True, False]
        ).iterrows():
            data = data.append(
                {
                    "setname": row["setname"],
                    "pvalue": row["empirical_pvalue"],
                    "type": "observed",
                    "value": row["observed"],
                    "boundary_max": row["observed"] + 0,
                    "boundary_in": row["observed"] - 0,
                    "err": 0,
                },
                ignore_index=True,
            )
            data = data.append(
                {
                    "setname": row["setname"],
                    "type": "null",
                    "pvalue": row["empirical_pvalue"],
                    "value": row["mean(null)"],
                    "boundary_max": row["mean(null)"] + row["var(null)"],
                    "boundary_min": row["mean(null)"] - row["var(null)"],
                    "err": row["var(null)"],
                },
                ignore_index=True,
            )

        data = data.reset_index(drop=True)
        data["ind"] = data.index

        g = sns.catplot(
            y="setname",
            x="value",
            hue="type",
            data=data,
            height=2.5,
            kind="bar",
            palette=palette_binary,
            orient="h",
            aspect=1.5,
        )

        k = 0
        col = ["black", "#b64d50"]
        max_obs = np.max(data[data["type"] == "observed"]["value"])
        for i, r in data[data["type"] == "observed"].iterrows():  # .sort_values(by=["pvalue"],ascending=True)
            pval = stars(r["pvalue"])
            g.facet_axis(0, 0).annotate(
                pval,
                xy=(r["value"] + 0.1 * r["value"],k + 0.25),
                color=col[r["pvalue"] < (0.1 * 2 / len(data))],
                fontsize=12,
            )
            k += 1

        g.facet_axis(0, 0).xaxis.grid(color="gray", linestyle="dotted", linewidth=0.5)
        g.set_xlabels(stat_name)
        #g.set_xticklabels(g.facet_axis(0, 0).get_xticklabels(), rotation=45, horizontalalignment='right')
        # g.set_ylabels("")
        g.despine(bottom=True, left=True)

    else:

        for i, row in table.sort_values(
            by=["empirical_pvalue", "observed"], ascending=[True, True]
        ).iterrows():
            data = data.append(
                {
                    "setname": row["setname"],
                    "pvalue": row["empirical_pvalue"],
                    "type": "observed",
                    "value": row["observed"],
                    "boundary_max": row["observed"] + 0,
                    "boundary_in": row["observed"] - 0,
                    "err": 0,
                },
                ignore_index=True,
            )

            data = data.append(
                {
                    "setname": row["setname"],
                    "type": "null",
                    "pvalue": row["empirical_pvalue"],
                    "value": row["mean(null)"],
                    "boundary_max": row["mean(null)"] + row["var(null)"],
                    "boundary_min": row["mean(null)"] - row["var(null)"],
                    "err": row["var(null)"],
                },
                ignore_index=True,
            )

        data = data.reset_index(drop=True)
        data["ind"] = data.index
        # data=data.iloc[0:20]

        # fig, ax = plt.subplots(1, 1, figsize=(10, 20), sharex=True,sharey=True)

        g = sns.catplot(
            y="value",
            x="setname",
            hue="type",
            data=data,
            height=2.5,
            kind="bar",
            palette=palette_binary,
            orient="v",
            aspect=1.5,
        )

        k = 0
        col = ["black", "#b64d50"]
        max_obs = np.max(data[data["type"] == "observed"]["value"])
        for i, r in data[data["type"] == "observed"].iterrows():  # .sort_values(by=["pvalue"],ascending=True)
            pval = stars(r["pvalue"])
            g.facet_axis(0, 0).annotate(
                "pval:" + pval,
                xy=(r["value"], k),
                xytext=(r["value"] + 0.05, k - 0.10),
                color=col[r["pvalue"] < (0.1 * 2 / len(data))],
                fontsize=12,
            )
            k += 1

        g.facet_axis(0, 0).xaxis.grid(color="gray", linestyle="dotted", linewidth=0.5)
        g.set_ylabels(stat_name)
        g.set_xticklabels(g.get_xticklabels(), rotation=45, horizontalalignment='right')
        # g.set_ylabels("")
        g.despine(bottom=True, left=True)

    if output_file.endswith('.pdf'):
        plt.savefig(output_file, format="pdf")
    elif output_file.endswith('.png'):
        plt.savefig(output_file, format="png")
    else:
        logging.warning('The null distribution figure can only be saved in pdf or png, forced to png')
        plt.savefig(output_file + '.png', format="png")


def paint_comparison_matrix(table_filename: 'pygna comparison output',
                            output_file: 'output figure file, specify png or pdf file',
                            rwr: 'use RWR is the table comes from a RWR analysis' = False,
                            single_geneset: 'use true if the comparison has been done for a single file' = False,
                            annotate: 'set true if uou want to print the pvalue inside the cell' = False):
    """
    This function plots the results of of a GNA test.
    Pass the results table generated by one of the functions and the output figure file (png or pdf).
    With RWR you can specify whether the test is a RWR association, in this case a different palette and limits are sets.
    Specify if the results are obtained using association with only one genesets (multiple setnames in the same file).
    Pass the annotate flag to have the pvalue annotation on the plot
    """

    # If RWR we want sequential, non divergent
    if rwr:
        palette = OrRd_9.mpl_colors
    else:
        palette = RdBu_11.mpl_colors

    with open(table_filename, "r") as f:
        table = pd.read_csv(f, sep=",")

    stat_name = table["analysis"][0]
    n_permutations = table["number_of_permutations"][0]

    # If we only have used one geneset, we need to add 1 col and 1 row
    # To have a symmetrical matrix for the pivot
    if single_geneset:
        table = table.loc[:, ['observed', 'setname_A', 'setname_B', 'empirical_pvalue']]
        for i in set(table['setname_A'].values.tolist()).union(set(table['setname_B'].values.tolist())):
            table = table.append({'setname_A': i, 'setname_B': i, 'observed': 0, 'empirical_pvalue': 1},
                                 ignore_index=True)

    pivot_table = table.pivot(values="observed", index="setname_A", columns="setname_B")
    pivot_table = pivot_table.fillna(0)

    # we make sure we have a triangual matrix (if I have A/B and not B/A
    # the pivot is leaving some cells NA )
    matrix = (pivot_table.T + pivot_table) - pivot_table.T * np.eye(len(pivot_table))
    if single_geneset:
        mask = np.triu(np.ones((len(pivot_table), len(pivot_table))))
    else:
        mask = np.ones((len(pivot_table), len(pivot_table))) - np.tril(np.ones((len(pivot_table), len(pivot_table))))

    # Create annotation from pvalues
    if annotate:
        annot = table.pivot(
            values="empirical_pvalue", index="setname_A", columns="setname_B"
        )
        annot = annot.fillna(0)
        annot = (annot.T.values + annot.values) - annot.T.values * np.eye(len(annot))
        func = np.vectorize(stars)
        annot = func(annot)
    else:
        annot = False

    fig, axes = plt.subplots(1, 1, figsize=(10, 10))
    fig.subplots_adjust(left=0.3, right=0.99, bottom=0.3, top=0.99)

    if rwr:
        print('plotting RWR')
        g2 = sns.heatmap(
            matrix,
            cmap=palette,
            ax=axes,
            square=True,
            xticklabels=1,
            yticklabels=1,
            mask=mask,
            annot=annot,
            cbar=True,
            linewidths=0.1,
            linecolor="white",
            fmt=""
        )
    else:
        g2 = sns.heatmap(
            matrix,
            ax=axes,
            cmap=palette,
            square=True,
            xticklabels=1,
            yticklabels=1,
            mask=mask,
            annot=annot,
            center=0,
            cbar=True,
            linewidths=0.1,
            linecolor="white",
            fmt="",
            annot_kws={"size": 15}
        )

    # Set the font size
    g2.set_yticklabels(g2.get_yticklabels(), rotation=0, fontsize=15)
    g2.set_xticklabels(g2.get_xticklabels(), rotation=90, fontsize=15)
    axes.set_xlabel("")
    axes.set_ylabel("")

    if output_file.endswith('.pdf'):
        plt.savefig(output_file, format="pdf")
    elif output_file.endswith('.png'):
        plt.savefig(output_file, format="png")
    else:
        logging.warning('The null distribution figure can only be saved in pdf or png, forced to png')
        fig.savefig(output_file + '.png', format="png")


def paint_volcano_plot(table_filename: 'pygna comparison output',
                       output_file: 'output figure file, specify png or pdf file',
                       rwr: 'use RWR is the table comes from a RWR analysis' = False,
                       id_col="setname_B",
                       threshold_x=0.0,
                       threshold_y=2.0,
                       annotate=False):
    """
    This function plots the results of a GNA test of association of a single geneset against multiple pathways.
    Pass the results table generated by one of the functions and the output figure file (png or pdf).
    From the results table, a multiple testing correction is applied and the results are those plotted.
    The defined threshold are for x: zscore and y: -log10(pvalue)
    """

    out.apply_multiple_testing_correction(table_filename, pval_col="empirical_pvalue", method="fdr_bh", threshold=0.1)

    with open(table_filename, "r") as f:
        df = pd.read_csv(f, sep=",")

    stat_name = df["analysis"][0]
    n_permutations = df["number_of_permutations"][0]

    # Normalise the plotting value
    df['zscore'] = (df['observed'] - df['mean(null)']) / np.sqrt(df['var(null)'].values)

    # If it's a shortest path, mirror the zscore
    loc = 2
    if not rwr:
        loc = 1

    # When pvalue==0 the -log would be infinite, hence we replace
    # pvalue with the permutation resolution - a tenth of the resolution
    # So that the plotted value is going to be 1 + max_log
    sig_th = 1 / n_permutations - 1 / n_permutations / 10
    df['bh_pvalue'] = df['bh_pvalue'] + (df['bh_pvalue'] == 0.0) * sig_th

    # transform in -log10(pvalue)
    df['-log10(p)'] = -np.log10(df['bh_pvalue'].values)

    pp.VolcanoPlot(df, output_file, p_col='-log10(p)', id_col=id_col, plotting_col="zscore", x_threshold=threshold_x,
                   y_threshold=threshold_y, y_label='-log10(pvalue)', x_label='z-score', annotate=annotate, loc=loc)


#######################################################
#################### EXTRA ############################
########################################################


def plot_adjacency(
    network: "network_filename",
    output_file: 'use png or pdf for output figure',
    clusters_file: "file of clusters to order the nodes" = None,
    size: "number of genes to plot, allows to cut the adjacency matrix" = None,
):
    """
    This function plots the adjacency matrix of a network. If a geneset file is passed, the matrix is organised by the different sets in the geneset.
    For the moment the genelist needs to be complete and non overlapping.
    This function has been mostly used for the generation of plots in the paper.
    """

    graph = rc.ReadTsv(network).get_data()
    if len(graph.nodes) > 1000:
        logging.warning("Graph is larger than 1k nodes, plotting might take too long")

    nodelist = None
    s = 0
    nodelabels = []
    if clusters_file:
        geneset = rc.ReadGmt(clusters_file).get_data()
        nodelist = [k for i, v in geneset.items() for k in v]
        for i, v in geneset.items():
            s += 1
            nodelabels.append([s] * len(v))
    nodelabels = [i for j in nodelabels for i in j]
    nodelabels = np.asarray(nodelabels)[np.newaxis, :]

    matrix = nx.adjacency_matrix(graph, nodelist=nodelist).toarray()

    if size:
        matrix = matrix[0: int(size), 0: int(size)]
        nodelabels = nodelabels[:, 0: int(size)]

    # Set font size
    sns.set(font_scale=2)
    if clusters_file:
        f, axes = plt.subplots(
            2,
            2,
            figsize=(10, 10),
            gridspec_kw={
                "width_ratios": [1, 100],
                "height_ratios": [1, 100],
                "wspace": 0.005,
                "hspace": 0.005,
            },
        )
        sns.heatmap(
            [[0, 0], [0, 0]],
            cmap="Greys",
            vmin=0,
            vmax=1,
            square=True,
            ax=axes[0, 0],
            cbar=False,
            xticklabels=False,
            yticklabels=False,
        )
        sns.heatmap(
            matrix,
            cmap="Greys",
            vmin=0,
            vmax=1,
            square=True,
            ax=axes[1, 1],
            cbar=False,
            xticklabels=False,
            yticklabels=False,
        )

        count = 0
        for i, v in geneset.items():
            if size and count < int(size):
                axes[0, 1].annotate(i, xy=(count, 0), xytext=(count, 0))
            elif size and count >= int(size):
                pass
            else:
                axes[0, 1].annotate(i, xy=(count, 0), xytext=(count, 0))
            count = count + len(v)

        g = sns.heatmap(
            nodelabels,
            xticklabels=False,
            yticklabels=False,
            vmin=1,
            vmax=len(geneset.keys()),
            square=False,
            cbar=False,
            linewidths=0,
            cmap="Set3",
            ax=axes[0, 1],
        )
        g = sns.heatmap(
            nodelabels.T,
            xticklabels=False,
            yticklabels=False,
            vmin=1,
            vmax=len(geneset.keys()),
            square=False,
            cbar=False,
            linewidths=0,
            cmap="Set3",
            ax=axes[1, 0],
        )
    else:
        f, axes = plt.subplots(1, 1, figsize=(10, 10))
        sns.heatmap(
            matrix,
            cmap="Greys",
            vmin=0,
            vmax=1,
            square=True,
            ax=axes[1, 1],
            cbar=False,
            xticklabels=False,
            yticklabels=False,
        )

    if output_file.endswith('.pdf'):
        plt.savefig(output_file, format="pdf")
    elif output_file.endswith('.png'):
        plt.savefig(output_file, format="png")
    else:
        logging.warning('The null distribution figure can only be saved in pdf or png, forced to png')
        f.savefig(output_file + '.png', format="png")


def stars(pvalue: float) -> str:
    """
    Converts a pvalue into the corresponding stars:

    pval>0.05 : ""

    0.01< pval <= 0.05 : *

    0.001< pval <=0.01: * *

    0.0001< pval <= 0.001: * * *

    pval <=0.0001: * * * *

    :param pvalue: The P-Value to be converted
    :return: A string corresponding to the significance of the P-Value

    Example
    _______
    >>> label = stars(0.027)
    """
    if not math.isnan(pvalue):
        if pvalue > 0.05:
            return ""
        elif 0.01 < pvalue <= 0.05:
            return "*"
        elif 0.001 < pvalue <= 0.01:
            return "**"
        elif 0.0001 < pvalue <= 0.001:
            return "***"
        else:
            return "****"


def paint_summary_gnt(output_figure: 'output figure filename',
                      *input_tables: 'pass all gnt output tables that need to be included in the plot',
                      setname: 'name of the dataset' = None,
                      threshold: 'Value to threshold the colors' = 0.05,
                      column_filter: 'column where the threshold is applied' = 'empirical_pvalue',
                      larger: 'if True the threshold is set as lower limit' = False,
                      less_tests: 'comma separated string of the tests that are significant if lower than expected, '
                                  'otherwise pass empty string' = 'topology_sp'
                      ):
    """
    This function plots the summary of a number of gnt test tables. It must be specified the name of the figure that is
    going to be used as output.
    """

    df = pd.DataFrame()

    for t in input_tables:
        # Open all the input tables
        with open(t, 'r') as f:
            tab = pd.read_csv(f)
            print(tab.head())
            # Retrieve only the specific setname
            if setname:
                tab = tab[tab['setname'] == setname]

            # If the tab is longer than one element the plot cannot be shown
            if len(tab) > 1:
                logging.warning('Showing multiple datasets')
            df = df.append(tab)

    ### All that come before can be decoupled (CLI/API)

    df['zscore'] = (df['observed'] - df['mean(null)']) / np.sqrt(df['var(null)'].values)

    if less_tests != '':
        ttt = less_tests.split(',')
        df['zscore'] = df['zscore'] * ((-1) * np.array([i in ttt for i in df['analysis']]) + 1 * np.array(
            [i not in ttt for i in df['analysis']]))

    if larger:
        df['significant'] = df[column_filter] > float(threshold)
    else:
        df['significant'] = df[column_filter] < float(threshold)

    df = df.sort_values(by=['setname', 'analysis'], ascending=False)

    # CHoose size of the plot depending on the rows
    height = len(df.analysis.unique())

    if setname:
        g1 = sns.catplot(x='zscore', y='analysis', data=df, hue='significant', kind='point', join=False,
                         hue_order=[False, True],
                         palette=RdBu_4.mpl_colors[0::3][::-1], height=1 + height / 2, sharex=False, aspect=1.5)
        g1.axes[0][0].xaxis.grid(True)
        g1.axes[0][0].yaxis.grid(True)
        g1.axes[0][0].spines['bottom'].set_visible(True)

    else:
        g1 = sns.catplot(x='zscore', y='analysis', data=df, hue='significant', kind='point', join=False,
                         hue_order=[False, True],
                         col='setname', col_wrap=2, palette=RdBu_4.mpl_colors[0::3][::-1], height=height / 2,
                         aspect=1.5, sharex=True)

        for a in g1.axes:
            a.xaxis.grid(True)
            a.yaxis.grid(True)
            a.spines['bottom'].set_visible(True)

    g1.fig.savefig(output_figure)



import logging
import networkx as nx
import numpy as np
import sys
import multiprocessing
from pygna.utils import get_sampling_p, get_node_bins_map

class StatisticalTest:
    """
    This class implements the statistical analysis performed by Pygna.
    It performs the statistical tests on the given network, elaborates the number of observed genes, the pvalue etc.
    Please refer to the single method documentation for the returning values
    """

    def __init__(self, test_statistic, network: nx.Graph, diz: dict = {}, degree_bins = 1):
        """
        :param test_statistic: the statistical function to be used for the calculation of the empirical p-value and the
        null distribution
        :param network: the network to be used for the analysis
        :param diz: the dictionary containing the genes
        """
        self.__test_statistic = test_statistic
        self.__network = network
        self.__diz = diz
        self.degree_bins = degree_bins
        self.node_2_bin_map = None
        self.bins = None

        if (type(self.__network) is nx.Graph) or (type(self.__network) is nx.DiGraph):
            self.__universe = list(set(self.__network.nodes()))
        elif type(self.__network) is dict:
            self.__universe = list(set(self.__network.keys()))
        else:
            logging.error("Unknown network type: %s" % type(self.__network))
            sys.exit(-1)

        # If number of bins is specified, the mapping is generated
        if (degree_bins>1):

            if type(self.__network) is dict:
                logging.error('Cannot do degree corrected sampling without the whole network')

            degree= list(self.__network.degree(nbunch = self.__universe))
            [nn,dd] = list(zip(*degree))

            # We make sure that later the universe corresponds to the mapping
            self.__universe = list(nn)
            histogram, bins_edge = np.histogram(list(dd), bins = degree_bins, density = False)
            self.bins = {int(i):{'nk':histogram[i], 'range':(bins_edge[i],bins_edge[i+1])} for i in range(len(histogram))}
            self.bins[len(self.bins.keys())-1]['range'] = (self.bins[len(self.bins.keys())-1]['range'][0], self.bins[len(self.bins.keys())-1]['range'][1]+1)
            self.node_2_bin_map = get_node_bins_map(degree, self.bins)


    def empirical_pvalue(
        self,
        geneset: set,
        alternative: str = "less",
        max_iter: int = 100,
        cores: int = 1,
    ) -> [int, float, float, int, int, int]:
        """
        Calculate the empirical pvalue on the genes list

        :param geneset: the geneset to elaborate
        :param alternative: the pvalue selection of the observed genes
        :param max_iter: the number of iterations to be performed
        :param cores: the number of cores to be used
        :return observed, pvalue, null_distribution, len(mapped_genesetA), len(mapped_genesetB): the list with the data calculated
        """
        # mapping geneset
        mapped_geneset = sorted(list(set(geneset).intersection(set(self.__universe))))
        if len(mapped_geneset) == 0:
            return 0, 1, np.array([0]), 0, 0
        else:
            logging.info(
                "Mapped %d genes out of %d." % (len(mapped_geneset), len(geneset))
            )

            sampling_p = None
            if self.degree_bins > 1:
                # if more than one bin is specified, the sampling probability is generated
                sampling_p = get_sampling_p(mapped_geneset, self.__network, self.bins, self.node_2_bin_map)

            observed = self.__test_statistic(
                self.__network, mapped_geneset, self.__diz, observed_flag=True
            )
            null_distribution = StatisticalTest.get_null_distribution_mp(
                self, mapped_geneset, max_iter, n_proc=cores, sampling_p = sampling_p
            )
            # computing empirical pvalue
            if alternative == "greater":
                pvalue = (np.sum(null_distribution >= observed) + 1) / (
                    float(max_iter) + 1
                )
            else:
                pvalue = (np.sum(null_distribution <= observed) + 1) / (
                    float(max_iter) + 1
                )

            return (
                observed,
                pvalue,
                null_distribution,
                len(mapped_geneset),
                len(geneset),
            )

    def get_null_distribution_mp(self, geneset: list, iter: int = 100, n_proc: int = 1, sampling_p = None):
        """
        Calculate the null distribution with multiple cores on the geneset

        :param geneset: the geneset to be used
        :param iter: the number of iterations to perform
        :param n_proc: the number of cpu to use for the elaboration
        :return: the array with null distribution
        """
        if n_proc == 1:
            null_distribution = StatisticalTest.get_null_distribution(
                self, geneset, iter, sampling_p
            )

        else:

            p = multiprocessing.Pool(n_proc)
            n_trial = int(iter / n_proc)
            results = [
                p.apply_async(
                    StatisticalTest.get_null_distribution, args=(self, geneset, n_trial, sampling_p)
                )
                for w in list(range(1, n_proc + 1))
            ]
            null_distribution = np.array([])
            for r in results:
                null_distribution = np.hstack((null_distribution, np.array(r.get())))
            p.close()

        return np.asarray(null_distribution)

    def get_null_distribution(self, geneset: list, n_samples: int, sampling_p = None):
        """
        Calculate the null distribution over the geneset

        :param geneset: the geneset to be used
        :param n_samples: the number of samples to be taken
        :return: the random distribution calculated
        """
        np.random.seed()
        random_dist = []
        for i in range(n_samples):
            random_sample = np.random.choice(
                self.__universe, len(geneset), replace=False, p = sampling_p
            )
            random_dist.append(
                self.__test_statistic(self.__network, set(random_sample), self.__diz)
            )

        return random_dist


###############################################################################
###  TEST STATISTICS FOR SINGLE GENESET  ######################################
###############################################################################


def geneset_localisation_statistic_median(
    network: nx.Graph, geneset: set, diz: dict = {}, observed_flag: bool = False
) -> float:
    """
    Calculate the median shortest path for each node

    :param network: the network used in the analysis
    :param geneset: the geneset to analyse
    :param diz: the dictionary containing the genes name
    :param observed_flag: whether the gene has been already observed
    """
    cum_sum = 0.0
    geneset_index = [diz["nodes"].index(i) for i in geneset]

    for u in geneset_index:
        d_uv = []
        for v in geneset_index:
            d_uv.append(diz["matrix"][v][u] + diz["matrix"][u][v])

        cum_sum += np.median(d_uv)
    return cum_sum / float(len(geneset))


def geneset_localisation_statistic(
    network: nx.Graph, geneset: set, diz: dict = {}, observed_flag: bool = False
) -> float:
    """
    Identify the genes in a geneset

    :param network: the network used in the analysis
    :param geneset: the geneset to analyse
    :param diz: the dictionary containing the genes name
    :param observed_flag: whether the gene has been already observed
    """
    n = np.array([diz["nodes"].index(i) for i in geneset])
    diz = diz["matrix"]

    sub_matrix = diz[n[:, None], n]
    min_columns = np.amin(sub_matrix, axis=0)
    sum_columns = np.sum(min_columns)
    return sum_columns / len(n)


def geneset_module_statistic(
    network: nx.Graph, geneset: set, diz: dict = {}, observed_flag: bool = False
) -> float:
    """
    Evaluate the length of a observed network

    :param network: the network used in the analysis
    :param geneset: the geneset to analyse
    :param diz: the dictionary containing the genes name
    :param observed_flag: whether the gene has been already observed
    """
    # Largest Connected Component for the subgraph induced by the geneset
    module = nx.subgraph(network, geneset)

    if observed_flag == True:
        pass
    cc = sorted(list(nx.connected_components(module)), key=len, reverse=True)

    if len(cc) > 0:
        return len(cc[0])
    else:
        return 0


def geneset_total_degree_statistic(
    network: nx.Graph, geneset: set, diz: dict = {}, observed_flag: bool = False
) -> float:
    """
    Total degree of the geneset: average total_degree

    :param network: the network used in the analysis
    :param geneset: the geneset to analyse
    :param diz: the dictionary containing the genes name
    :param observed_flag: whether the gene has been already observed
    """
    degree = nx.degree(network)
    geneset = list(geneset)
    total = np.array([degree[g] for g in geneset])
    return np.average(total)


def geneset_internal_degree_statistic(
    network: nx.Graph, geneset: set, diz: dict = {}, observed_flag: bool = False
) -> float:
    """
    Internal degree ratio: average of the ratio internal_degree/total_degree

    :param network: the network used in the analysis
    :param geneset: the geneset to analyse
    :param diz: the dictionary containing the genes name
    :param observed_flag: whether the gene has been already observed
    """
    degree = nx.degree(network)
    total = np.array([degree[g] for g in geneset])

    subgraph = network.subgraph(geneset)
    degree_internal = nx.degree(subgraph)
    internal = np.array([degree_internal[g] for g in geneset])

    ratio = internal / total
    ratio[total == 0] = 0.5

    return np.average(ratio)


def geneset_RW_statistic(
    network: nx.Graph, geneset: set, diz: dict = {}, observed_flag: bool = False
) -> np.ndarray:
    """
    Poisson binomial probability, sum of interaction probabilities for the genes in the geneset

    :param network: the network used in the analysis
    :param geneset: the geneset to analyse
    :param diz: the dictionary containing the genes name
    :param observed_flag: whether the gene has been already observed
    """
    try:
        diz["matrix"]
    except KeyError:
        print("The dictionary doesnt have a matrix key")
        raise

    geneset_index = [diz["nodes"].index(i) for i in geneset]
    prob = [diz["matrix"][i, j] for i in geneset_index for j in geneset_index if i != j]
    prob = np.sum(prob)
    return prob



import abc

import pandas as pd
import tables
import networkx as nx
import numpy as np
import logging
import sys


class ReadingData(object):
    """
    Abstract class used to read different types of file. You can implement your own reading method, but remember
    that each subclass must implement the **readfile** and **get_data** methods
    """

    def __init__(self):
        super(ReadingData, self).__init__()

    @abc.abstractmethod
    def __readfile(self):
        """
        Implement this method to read your custom file
        """
        raise NotImplementedError

    @abc.abstractmethod
    def get_data(self):
        """
        Get the data from the reading class. This method must be always overridden
        """
        raise NotImplementedError


class ReadTsv(ReadingData):
    """
    This class is used to read and parse a network file in a tab-separated format (tsv).
    """

    def __init__(self, filename: str, pd_table: bool = False, int_type: int = None):
        """
        :param filename: represents the path to the network file
        :param pd_table: if the results is going to be a pd.dataframe
        """
        super().__init__()
        self.filename = filename
        self.int_type = int_type
        self.pd_table = pd_table
        self.interactions = []

        if not self.pd_table:
            self.__readfile()
        self.graph = self._convert_to_graph()

    def __readfile(self) -> None:
        """
        This method read the file and saves the data inside a class attribute
        """
        with open(self.filename, "r") as f:
            for record in f:
                if record.startswith("#"):
                    continue

                fields = record.strip().split("\t")
                if self.int_type:
                    types = fields[3].split(";")
                    if self.int_type in types:
                        self.interactions.append((fields[0], fields[1]))
                    else:
                        continue
                else:
                    self.interactions.append((fields[0], fields[1]))

    def _convert_to_graph(self) -> nx.Graph:
        """
        Converts the interactions into a graph object

        :return: graph from the interactions
        """
        graph = nx.Graph()
        graph.add_edges_from(self.interactions)
        graph.remove_edges_from(graph.selfloop_edges())
        return graph

    def get_data(self) -> pd.DataFrame or list:
        """
        Returns the data of the tsv file

        :return: list representing the genes read in the file

        Example
        _______
        >>> tsvdata = ReadTsv("mydata.tsv").get_data()
        """
        if self.pd_table:
            return pd.read_table(self.filename)
        else:
            return self.interactions

    def get_network(self) -> nx.Graph:
        """
        Returns the nx.graph object of the network

        :return: graph containing the network information

        Example
        _______
        >>> tsvdata = ReadTsv("mydata.tsv").get_network()
        """
        return self.graph


class ReadGmt(ReadingData):
    """
    This class is used to read a gmt file, which contains information about the genes with a setname and separated by a comma
    """

    def __init__(self, filename: str, read_descriptor: bool = False):
        """
        :param filename: represents the path to the geneset file
        :param read_descriptor: if the descriptor is given. Default = False
        """
        super().__init__()
        self.filename = filename
        self.read_descriptor = read_descriptor

        self.gmt_data = self.__readfile()

    def __readfile(self) -> dict:
        """
        This method reads the geneset file into a variable

        :return: gene_list representing the list of genes
        """
        gene_lists = dict()
        with open(self.filename, "r") as f:
            for record in f:
                fields = record.strip().split("\t")
                if self.read_descriptor:
                    gene_lists[fields[0]] = {}
                    gene_lists[fields[0]]["genes"] = fields[2:]
                    gene_lists[fields[0]]["descriptor"] = fields[1]
                else:
                    gene_lists[fields[0]] = fields[2:]
            return gene_lists

    def get_data(self) -> dict:
        """
        Returns the data of the gmt file

        :return: dict representing the genes list

        Example
        _______
        >>> gmtdata = ReadGmt("mydata.gmt").get_data()
        """
        return self.gmt_data

    def get_geneset(self, setname: str = None) -> dict:
        """
        Returns the geneset from the gmt file

        :param setname: the setname to extract
        :return: the geneset data

        Example
        _______
        >>> gmtdata = ReadGmt("mydata.gmt").get_geneset("brca")
        """
        if setname is not None:
            if setname in self.gmt_data:
                temp = self.gmt_data[setname]
                self.gmt_data.clear()
                self.gmt_data[setname] = temp
            else:
                logging.error("Cannot find geneset: %s" % setname)
                sys.exit(-1)
        return self.gmt_data


class ReadCsv(ReadingData):
    """
    This class is used to read a csv file.
    """

    def __init__(self, filename: str, sep: str = ",", use_cols: list = None, column_to_fill: str = None):
        """
        :param filename: represents the path to the data file
        :param sep: the separator to be used
        :param use_cols: columns used to be read and grouped
        :param column_to_fill: column to fill the NA values
        """
        super().__init__()
        self.filename = filename
        self.sep = sep
        self.use_cols = use_cols
        self.name_column = column_to_fill

        self.data = self.__readfile()
        if self.name_column is not None:
            self._fill_na_column()

    def __readfile(self) -> pd.DataFrame:
        """
        This method read the file and saves the data inside a class attribute

        :return: dataframe representing teh data read inside the .csv
        """
        with open(self.filename, "r") as f:
            table = pd.read_csv(f, sep=self.sep, usecols=self.use_cols)
            return table

    def get_data(self) -> pd.DataFrame:
        """
        Returns the data of the csv file

        :return: dataframe representing the data read inside the .csv

        Example
        _______
        >>> csvdata = ReadCsv("mydata.csv").get_data()
        """
        return self.data

    def _fill_na_column(self) -> None:
        """
        Fill the N/A values with a (str) 0
        """
        self.data[self.name_column].fillna(0, inplace=True)
        self.data[self.name_column] = self.data[self.name_column].astype(int)


class ReadTxt(ReadingData):
    """
    This class reads a txt file containing a single gene per line
    """

    def __init__(self, filename: str):
        super().__init__()
        self.filename = filename
        self.data = []
        self.__readfile()

    def __readfile(self) -> None:
        """
        Read the file, line per line
        """
        with open(self.filename, "r") as f:
            gene_line = f.readline()
            while gene_line:
                self.data.append(gene_line)
                gene_line = f.readline()

    def get_data(self) -> pd.DataFrame:
        """
        Get the dataframe from the class

        :return: dataframe object from the file read

        Example
        _______
        >>> txtdata = ReadTxt("mydata.txt").get_data()
        """
        return pd.DataFrame(self.data)


class ReadDistanceMatrix(ReadingData):
    """
    This class read a distance matrix in the HDF5 format
    """

    def __init__(self, filename: str, in_memory: bool = False):
        """
        :param filename: the path of the file to be read
        :param in_memory: keep the matrix in memory or not
        """
        super().__init__()
        self.filename = filename
        self.nodes = None
        self.data = None
        self.memory = in_memory

        self.__readfile()
        if type(self.nodes[0]) == bytes:
            self._decode()

    def __readfile(self) -> None:
        """
        This method read and stores matrix information in memory or by reading it on the disk
        """
        if self.memory:
            hdf5_file = tables.open_file(self.filename, mode="r", driver="H5FD_CORE")
        else:
            hdf5_file = tables.open_file(self.filename, mode="r")
        self.nodes = list(hdf5_file.root.nodes[:])
        self.data = hdf5_file.root.matrix[:]

    def _decode(self) -> None:
        """
        Elaborate teh nodes from the graph
        """
        self.nodes = [i.decode() for i in self.nodes]

    def get_data(self) -> [list, np.matrix]:
        """Return the data of the HDF5 Matrix

        :return: table data, the data of the HDF5 Matrix and table nodes, the nodes of the HDF5 Matrix

        Example
        _______
        >>> nodes, data = ReadDistanceMatrix("mydata.hdf5").get_data()
        """
        return self.nodes, self.data



import os
from setuptools import find_packages, setup

# determining the directory containing setup.py
setup_path = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(setup_path, 'README.md'), encoding='utf-8') as f:
    readme = f.read()

setup(
    # package information
    name='pygna',
    packages=find_packages(),
    version='3.3.1',
    description='Geneset Network Analysis',
    long_description=readme,
    license='MIT',
    url='https://github.com/stracquadaniolab/pygna',
    keywords='Bioinformatics Network Statistics',

    #  author information
    author='Viola Fanfani, Giovanni Stracquadanio',
    author_email='v.fanfani@sms.ed.ac.uk',

    # installation info and requirements
    install_requires=[
        'pandas',
        'numpy',
        'scipy',
        'matplotlib',
        'pyyaml',
        'tables>=3.4.4',
        'seaborn>=0.9',
        'palettable',
        'networkx==2.3',
        'statsmodels',
        'argh',
        'mygene'
    ],
    setup_requires=[],

    # test info and requirements
    test_suite='tests',
    tests_require=[],
    python_requires='>=3.8',

    # package deployment info
    include_package_data=True,
    zip_safe=False,

    # all tools have cli interface
    entry_points={
        'console_scripts': [
            'pygna=pygna.cli:main',
        ],
    },

    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.8',
        'Topic :: Scientific/Engineering :: Bio-Informatics'
    ]
)



import matplotlib.pyplot as plt
import seaborn as sns
import logging
import numpy as np
from matplotlib.offsetbox import AnchoredText
import networkx as nx


def plot_degree(degree_object: nx.Graph, output_file: str):
    """
    Diagnosis tool for the degree object. It takes a graph and plot its statistics

    :param degree_object: the graph to plot
    :param output_file: the path to save the file

    Example
    _______
    >>> graph = nx.complete_graph(100)
    >>> plot_degree(nx.degree(graph), "graph_degree")
    """

    D = dict(degree_object)
    degrees = {k: v for k, v in D.items()}
    degree_values = np.array(list(degrees.values()))

    fig, axes = plt.subplots(1, figsize=(10, 10))
    g1 = sns.distplot(degree_values, hist=True, ax=axes)

    key_max = max(degrees.keys(), key=(lambda k: degrees[k]))
    g1 = sns.distplot([degrees[key_max]], hist=False, kde=False, rug=True, color='r', ax=axes)
    axes.annotate('%s: %d' % (key_max, degrees[key_max]), xy=(degrees[key_max], 0),
                  xytext=(degrees[key_max], axes.dataLim.y1 / 2), arrowprops=dict(arrowstyle="->"))

    g1 = sns.distplot([np.median(degree_values)], hist=False, kde=False, rug=True, color='r', ax=axes)
    axes.annotate('median %f' % np.median(degree_values), xy=(np.median(degree_values), 0),
                  xytext=(np.median(degree_values), axes.dataLim.y1 / 2), arrowprops=dict(arrowstyle="->"))

    sns.despine(ax=axes, top=True, bottom=False, right=True, left=True)
    g1.set_ylabel("Density")
    g1.set_xlabel("Node Degree")

    if output_file.endswith('.pdf'):
        plt.savefig(output_file, format="pdf")
    elif output_file.endswith('.png'):
        plt.savefig(output_file, format="png")
    else:
        logging.warning('The null distribution figure can only be saved in pdf or png, forced to png')
        fig.savefig(output_file + '.png', format="png")


def plot_connected_components(c_components: nx.connected_components, output_file: str) -> None:
    """
    Diagnosis tool for the connected components object.
    Creates the histogram of the components length, to analyse the relationship between the lcc
    and the other c_components, and prints some overall stats about the connected components

    :param c_components: the list of the connected components
    :param output_file:  the path to save the file

    Example
    _______
    >>> from pygna import diagnostic
    >>> import pygna.reading_class as rc
    >>> network = rc.ReadTsv("network_file.tsv").get_network()
    >>> geneset = rc.ReadGmt("geneset_input_file.gmt").get_geneset("brca")
    >>> c_components_figure_file = "components_plot.pdf"
    >>> for setname, item in geneset.items():
    ...     graph = nx.subgraph(network, item)
    ...     diagnostic.plot_connected_components(nx.connected_components(graph), c_components_figure_file)


    """
    c_components_len = [len(k) for k in c_components]

    logging.info("Number of cc %d" % (len(c_components_len)))
    logging.info("First five cc" + str(c_components_len[0:5]))
    logging.info("Mean length of cc %d" % (np.mean(c_components_len)))

    fig, axes = plt.subplots(1, figsize=(10, 10))
    g1 = sns.distplot(c_components_len, hist=True, kde=False, ax=axes, norm_hist=False)
    g1 = sns.distplot([np.max(c_components_len)], hist=False, kde=False, rug=True, color='r', ax=axes, norm_hist=False)

    axes.annotate('LCC: %d' % np.max(c_components_len), xy=(np.max(c_components_len), 0),
                  xytext=(np.max(c_components_len) - 10, axes.dataLim.y1 / 4), arrowprops=dict(arrowstyle="->"))

    sns.despine(ax=axes, top=True, bottom=False, right=True, left=True)
    g1.set_ylabel("Number of CC")
    g1.set_xlabel("Size of CC")

    if output_file.endswith('.pdf'):
        plt.savefig(output_file, format="pdf")
    elif output_file.endswith('.png'):
        plt.savefig(output_file, format="png")
    else:
        logging.warning('The null distribution figure can only be saved in pdf or png, forced to png')
        fig.savefig(output_file + '.png', format="png")


def plot_diffusion_matrix(nodes: list, matrix: np.ndarray, filename: str, show_labels: bool = False) -> None:
    """
    Diagnosis tool for a diffusion matrix. It shows the weighted adjacency matrix that is the output of a build process

    :param nodes: the network nodes
    :param matrix: the diffusion matrix
    :param filename: the path to save the file
    :param show_labels: if labels should be plotted

    Example
    _______
    >>> nodes = ["A", "B", "C"]
    >>> matrix = np.random.rand(3,2)
    >>> plot_diffusion_matrix(nodes, matrix, "diff_matrix.pdf")
    """

    logging.info("Plotting figure as " + str(filename))
    fig, axes = plt.subplots(1)
    axes.imshow(matrix, cmap="PuBu")
    if show_labels:
        pass
    plt.show()
    fig.savefig(filename + ".pdf", format="pdf")


def plot_null_distribution(null_distribution: list, observed: list, output_file: str, setname: str,
                           alternative: str = "greater") -> None:
    """
    Saves the density plot of the null distribution and pinpoints the observed value. Please refer to the CLI for the usage of this function

    :param null_distribution: the list with the values from the null distribution
    :param observed: list of the observed genes
    :param output_file: the path to save the file
    :param setname: the name of the gene set
    :param alternative: use "greater" if you want to take the genes with greater than the observed value

    Example
    _______
    >>> from pygna import diagnostic
    >>> import pygna.statistical_test as st
    >>> import pygna.reading_class as rc
    >>> network = rc.ReadTsv("network_file.tsv").get_network()
    >>> st_test = st.StatisticalTest(st.geneset_total_degree_statistic, network)
    >>> geneset = rc.ReadGmt("geneset_file.gmt").get_geneset("brca")
    >>> for setname, item in geneset.items():
    ...     observed, pvalue, null_d, n_mapped, n_geneset = st_test.empirical_pvalue(item, max_iter=500, alternative="greater", cores=10)
    ...     diagnostic.plot_null_distribution(null_d, observed, "./results/" + setname +'_total_degree_null_distribution.pdf', setname=setname)
    """

    fig, axes = plt.subplots(1, figsize=(8, 6))
    g1 = sns.distplot(null_distribution, hist=True, kde=True, rug=False, ax=axes)
    if alternative == "greater":
        if len(null_distribution[null_distribution > observed]):
            g3 = sns.distplot(null_distribution[null_distribution > observed], hist=False, kde=False, rug=True,
                              rug_kws={'height': 1 / 50}, color="r", ax=axes)
    else:
        if len(null_distribution[null_distribution < observed]):
            g3 = sns.distplot(null_distribution[null_distribution < observed], hist=False, kde=False, rug=True,
                              rug_kws={'height': 1 / 50}, color="r", ax=axes)
    ymax = axes.dataLim.y1
    xmax = axes.dataLim.x1
    print('xmax %f' % xmax)
    g4 = axes.stem([observed], [ymax / 2], "r", "r--")

    sns.despine(ax=axes, top=True, bottom=False, right=True, left=True)
    anchored_text = AnchoredText("Observed:%1.1E" % observed, loc=1, prop={'fontsize': 18, 'color': 'r'},
                                 **{'frameon': False})
    axes.add_artist(anchored_text)
    axes.set_xlabel('Statistics', fontsize=18)
    axes.set_ylabel('Density', fontsize=18)
    logging.info("Output for diagnostic null distribution: " + output_file)
    if output_file.endswith('.pdf'):
        fig.savefig(output_file, format="pdf")
    elif output_file.endswith('.png'):
        fig.savefig(output_file, format="png")
    else:
        logging.warning('The null distribution figure can only be saved in pdf or png, forced to png')
        fig.savefig(output_file + '.png', format="png")



import logging
import pandas as pd


class TableElaboration:
    """
    This class contains static methods to clean and filter specific columns of a table
    """

    @staticmethod
    def clean_table(table: pd.DataFrame, stat_col: str = "stat") -> pd.DataFrame:
        """
        This function clean the table from the N/A values

        :param table: dataframerepresenting the table to be cleaned
        :param stat_col: the column to be cleaned
        :return: the table cleaned from the N/A values

        Example
        _______
        >>> import numpy as np
        >>> table = pd.DataFrame(np.random.randint(0,100,size=(100, 1)), columns=list('mycol'))
        >>> table = TableElaboration.clean_table(table, "mycol")
        """
        logging.info("original table has %d rows" % len(table))
        table = table.dropna(subset=[stat_col])
        logging.info("cleaned table has %d rows" % len(table))
        return table

    @staticmethod
    def filter_table(table: pd.DataFrame, filter_column: str = "padj", alternative: str = "less",
                     threshold: float = 0.01) -> pd.DataFrame:
        """
        This method filters a table according to a filter rule èassed as input

        :param table: The table to be filtered
        :param filter_column: Column with the values to be filtered
        :param alternative: Alternative to use for the filterK with "less" the filter is applied <threshold; otherwise >= threshold
        :param threshold: Threshold for the filter
        :return: The table filtered

        Example
        _______
        >>> import numpy as np
        >>> table = pd.DataFrame(np.random.randint(0,100,size=(100, 1)), columns=list('pval'))
        >>> table = TableElaboration.filter_table(table, filter_column="pval")
        """
        threshold = float(threshold)
        try:
            if alternative == "less":
                table = table[table[filter_column] < threshold]
            else:
                table = table[table[filter_column] >= threshold]
        except Exception as e:
            logging.error("type error: " + str(e))
        return table
